@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006,2007,2008
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -113,12 +113,6 @@ mst_Boolean _gst_skip_compilation = false;
    Most often, the caller does not care about the returned value,
    since it often is called from a radically different context.  */
 OOP _gst_last_returned_value = NULL;
-
-/* This is set to true by the parser or the compiler if an error
-   (respectively, a parse error or a semantic error) is found, and
-   avoids that _gst_execute_statements tries to execute the result of
-   the compilation.  */
-mst_Boolean _gst_had_error = false;
 
 
 
@@ -206,12 +200,6 @@ static mst_Boolean compile_statements (tree_node statementList,
    them into a symbol OOP and returns that symbol.  */
 static OOP compute_selector (tree_node selectorExpr);
 
-/* Given CONSTEXPR, a section of the syntax tree that represents a
-   Smalltalk constant, this routine creates and returns an OOP to be
-   stored as a method literal in the method that's currently being
-   compiled.  */
-static OOP make_constant_oop (tree_node constExpr);
-
 /* Creates a new Array object that contains the literals for the
    method that's being compiled and returns it.  As a side effect, the
    currently allocated working literal vector is freed.  If there were
@@ -403,9 +391,6 @@ static void install_method (OOP methodOOP);
    _gst_prepare_execution_environment.  */
 static OOP termination_method;
 
-/* Used to number the lines starting from 1.  */
-static int line_offset;
-
 /* Used to abort really losing compiles, jumps back to the top level
    of the compiler */
 static jmp_buf bad_method;
@@ -478,12 +463,12 @@ _gst_install_initial_methods (void)
   install_method (termination_method);
 
   methodsForString = "\
-methodsFor: aCategoryString \
+methodsFor: aCategoryString [\
     \"Calling this method prepares the parser to receive methods \
       to be compiled and installed in the receiver's method dictionary. \
       The methods are put in the category identified by the parameter.\" \
     <primitive: VMpr_Behavior_methodsFor> \
-";
+]";
   _gst_set_compilation_class (_gst_behavior_class);
   _gst_set_compilation_category (_gst_string_new ("compiling methods"));
   _gst_push_smalltalk_string (_gst_string_new (methodsForString));
@@ -668,7 +653,7 @@ _gst_execute_statements (tree_node temps,
   methodOOP =
     _gst_compile_method (_gst_make_method (&statements->location, &loc,
 					   messagePattern, temps, NULL,
-					   statements),
+					   statements, false),
 			 true, false);
 
   SET_CLASS_ENVIRONMENT (_gst_undefined_object_class,
@@ -799,14 +784,13 @@ _gst_compile_method (tree_node method,
   int primitiveIndex;
   int stack_depth;
   inc_ptr incPtr;
+  gst_compiled_method compiledMethod;
 
   dup_message_receiver = false;
   literal_vec_curr = literal_vec;
   this_method_category = _gst_this_category;
   _gst_unregister_oop (_gst_latest_compiled_method);
   _gst_latest_compiled_method = _gst_nil_oop;
-  _gst_line_number (-1, true);
-  line_offset = method->location.first_line - 1;
 
   incPtr = INC_SAVE_POINTER ();
 
@@ -814,6 +798,11 @@ _gst_compile_method (tree_node method,
   _gst_push_new_scope ();
   inside_block = 0;
   selector = compute_selector (method->v_method.selectorExpr);
+  if (method->location.file_offset != -1 && _gst_get_cur_stream_prompt ())
+    _gst_line_number (method->location.first_line, LN_RESET);
+  else
+    _gst_line_number (-1, LN_RESET);
+
   INC_ADD_OOP (selector);
 
   if (_gst_declare_tracing)
@@ -906,6 +895,9 @@ _gst_compile_method (tree_node method,
 					_gst_this_category,
 					method->location.file_offset,
 					method->v_method.endPos);
+
+      compiledMethod = (gst_compiled_method) OOP_TO_OBJ (methodOOP);
+      compiledMethod->header.isOldSyntax = method->v_method.isOldSyntax;
       INC_ADD_OOP (methodOOP);
 
       if (install)
@@ -969,7 +961,7 @@ compile_expression (tree_node expr)
 void
 compile_simple_expression (tree_node expr)
 {
-  _gst_line_number (expr->location.first_line - line_offset, false);
+  _gst_line_number (expr->location.first_line, 0);
   switch (expr->nodeType)
     {
     case TREE_VARIABLE_NODE:
@@ -1090,7 +1082,7 @@ compile_constant (tree_node constExpr)
      bytecode, or add it to the literals.  */
   if (index == -1)
     {
-      constantOOP = make_constant_oop (constExpr);
+      constantOOP = _gst_make_constant_oop (constExpr);
       if (IS_INT (constantOOP))
         {
           intVal = TO_INT (constantOOP);
@@ -1101,8 +1093,6 @@ compile_constant (tree_node constExpr)
 	      return;
 	    }
         }
-      else
-        MAKE_OOP_READONLY (constantOOP, true);
 
       index = add_literal (constantOOP);
     }
@@ -1200,7 +1190,7 @@ compile_statements (tree_node statementList,
 
   if (isBlock)
     {
-      _gst_line_number (statementList->location.first_line - line_offset, true);
+      _gst_line_number (statementList->location.first_line, LN_FORCE);
       inside_block++;
     }
 
@@ -1225,7 +1215,7 @@ compile_statements (tree_node statementList,
 
   if (isBlock)
     {
-      _gst_line_number (-1, false);
+      _gst_line_number (-1, 0);
       inside_block--;
     }
 
@@ -1912,7 +1902,7 @@ compile_assignments (tree_node varList)
     {
       tree_node varName = varList->v_list.value;
 
-      _gst_line_number (varList->location.first_line - line_offset, false);
+      _gst_line_number (varList->location.first_line, 0);
       if (!_gst_find_variable (&variable, varName))
 	{
           if (varName->v_list.next)
@@ -1956,7 +1946,14 @@ compile_assignments (tree_node varList)
 	_gst_compile_byte (STORE_RECEIVER_VARIABLE, variable.varIndex);
 
       else
-	_gst_compile_byte (STORE_LIT_VARIABLE, variable.varIndex);
+	{
+	  /* This can become a message send, which might not return the
+	     value.  Compile it in a way that can be easily peephole
+	     optimized. */
+	  _gst_compile_byte (STORE_LIT_VARIABLE, variable.varIndex);
+	  _gst_compile_byte (POP_STACK_TOP, 0);
+	  _gst_compile_byte (PUSH_LIT_VARIABLE, variable.varIndex);
+	}
     }
 }
 
@@ -2061,13 +2058,39 @@ equal_constant (OOP oop,
 	}
       break;
 
-    case CONST_OOP:
-      if (oop == constExpr->v_const.val.oopVal)
-	return (true);
+    case CONST_DEFERRED_BINDING:
+      if (IS_OOP (oop) && OOP_CLASS (oop) == _gst_deferred_variable_binding_class)
+	{
+	  gst_deferred_variable_binding binding =
+	    (gst_deferred_variable_binding) OOP_TO_OBJ (oop);
+	  gst_object path = OOP_TO_OBJ (binding->path);
+	  int i, size = NUM_OOPS (path);
+	  OOP *pKey;
+	  tree_node varNode = constExpr->v_const.val.aVal;
+
+	  /* Use <= because we test the key first.  */
+	  for (i = 0, pKey = &binding->key; i <= size; pKey = &path->data[i++])
+	    {
+	      if (!varNode
+		  || *pKey != _gst_intern_string (varNode->v_list.name))
+		return (false);
+
+	      varNode = varNode->v_list.next;
+	      pKey = &path->data[i];
+	    }
+	}
       break;
 
     case CONST_BINDING:
-      if (oop == _gst_find_variable_binding (constExpr->v_const.val.aVal, false))
+      constExpr = _gst_find_variable_binding (constExpr->v_const.val.aVal);
+      if (!constExpr)
+	return (false);
+
+      assert (constExpr->v_const.constType != CONST_BINDING);
+      return equal_constant (oop, constExpr);
+
+    case CONST_OOP:
+      if (oop == constExpr->v_const.val.oopVal)
 	return (true);
       break;
 
@@ -2099,9 +2122,9 @@ equal_constant (OOP oop,
 }
 
 OOP
-make_constant_oop (tree_node constExpr)
+_gst_make_constant_oop (tree_node constExpr)
 {
-  tree_node arrayElt;
+  tree_node subexpr;
   int len, i;
   OOP resultOOP, elementOOP;
   inc_ptr incPtr;
@@ -2117,8 +2140,8 @@ make_constant_oop (tree_node constExpr)
 
   else if (constExpr->nodeType == TREE_ARRAY_ELT_LIST)
     {
-      for (len = 0, arrayElt = constExpr; arrayElt;
-	   len++, arrayElt = arrayElt->v_list.next);
+      for (len = 0, subexpr = constExpr; subexpr;
+	   len++, subexpr = subexpr->v_list.next);
 
       incPtr = INC_SAVE_POINTER ();
 
@@ -2128,15 +2151,14 @@ make_constant_oop (tree_node constExpr)
       instantiate_with (_gst_array_class, len, &resultOOP);
       INC_ADD_OOP (resultOOP);
 
-      for (i = 0, arrayElt = constExpr; i < len;
-	   i++, arrayElt = arrayElt->v_list.next)
+      for (i = 0, subexpr = constExpr; i < len;
+	   i++, subexpr = subexpr->v_list.next)
 	{
-	  elementOOP = make_constant_oop (arrayElt->v_list.value);
+	  elementOOP = _gst_make_constant_oop (subexpr->v_list.value);
 	  result = OOP_TO_OBJ (resultOOP);
 	  result->data[i] = elementOOP;
-	  if (IS_OOP (elementOOP))
-	    MAKE_OOP_READONLY (elementOOP, true);
 	}
+      MAKE_OOP_READONLY (resultOOP, true);
       INC_RESTORE_POINTER (incPtr);
       return (resultOOP);
     }
@@ -2159,47 +2181,80 @@ make_constant_oop (tree_node constExpr)
       return (floatq_new (constExpr->v_const.val.fVal));
 
     case CONST_STRING:
-      return (_gst_string_new (constExpr->v_const.val.sVal));
-
-    case CONST_OOP:
-      return (constExpr->v_const.val.oopVal);
+      resultOOP = _gst_string_new (constExpr->v_const.val.sVal);
+      MAKE_OOP_READONLY (resultOOP, true);
+      return (resultOOP);
 
     case CONST_BYTE_OBJECT:
       bo = constExpr->v_const.val.boVal;
       result = instantiate_with (bo->class, bo->size, &resultOOP);
       memcpy (result->data, bo->body, bo->size);
+      MAKE_OOP_READONLY (resultOOP, true);
       return (resultOOP);
 
+    case CONST_DEFERRED_BINDING:
+      {
+	gst_deferred_variable_binding dvb;
+	tree_node varNode = constExpr->v_const.val.aVal;
+
+        incPtr = INC_SAVE_POINTER ();
+        dvb = (gst_deferred_variable_binding)
+	  instantiate (_gst_deferred_variable_binding_class, &resultOOP);
+        INC_ADD_OOP (resultOOP);
+
+	dvb->key = _gst_intern_string (varNode->v_list.name);
+	dvb->class = _gst_this_class;
+	dvb->defaultDictionary = _gst_get_undeclared_dictionary ();
+	dvb->association = _gst_nil_oop;
+
+	varNode = varNode->v_list.next;
+	if (varNode)
+	  {
+	    int i, size = list_length (varNode);
+	    OOP arrayOOP;
+	    gst_object array =
+	      instantiate_with (_gst_array_class, size, &arrayOOP);
+
+	    dvb->path = arrayOOP;
+	    for (i = 0; i < size; i++, varNode = varNode->v_list.next)
+	      array->data[i] = _gst_intern_string (varNode->v_list.name);
+	  }
+
+        return (resultOOP);
+      }
+
     case CONST_BINDING:
-      resultOOP = _gst_find_variable_binding (constExpr->v_const.val.aVal,
-					      false);
-      if (IS_NIL (resultOOP))
+      subexpr = _gst_find_variable_binding (constExpr->v_const.val.aVal);
+      if (!subexpr)
 	{
 	  _gst_errorf_at (constExpr->location.first_line,
 			  "invalid variable binding");
           EXIT_COMPILATION ();
 	}
 
-      return (resultOOP);
+      assert (subexpr->v_const.constType != CONST_BINDING);
+      return _gst_make_constant_oop (subexpr);
+
+    case CONST_OOP:
+      return (constExpr->v_const.val.oopVal);
 
     case CONST_ARRAY:
-      for (len = 0, arrayElt = constExpr->v_const.val.aVal; arrayElt;
-	   len++, arrayElt = arrayElt->v_list.next);
+      for (len = 0, subexpr = constExpr->v_const.val.aVal; subexpr;
+	   len++, subexpr = subexpr->v_list.next);
 
       incPtr = INC_SAVE_POINTER ();
       result = instantiate_with (_gst_array_class, len, &resultOOP);
       INC_ADD_OOP (resultOOP);
 
-      for (i = 0, arrayElt = constExpr->v_const.val.aVal; i < len;
-	   i++, arrayElt = arrayElt->v_list.next)
+      for (i = 0, subexpr = constExpr->v_const.val.aVal; i < len;
+	   i++, subexpr = subexpr->v_list.next)
 	{
-	  elementOOP = make_constant_oop (arrayElt->v_list.value);
+	  elementOOP = _gst_make_constant_oop (subexpr->v_list.value);
 	  result = OOP_TO_OBJ (resultOOP);
 	  result->data[i] = elementOOP;
-	  if (IS_OOP (elementOOP))
-	    MAKE_OOP_READONLY (elementOOP, true);
 	}
 
+      MAKE_OOP_READONLY (resultOOP, true);
       INC_RESTORE_POINTER (incPtr);
       return (resultOOP);
     }
@@ -2315,20 +2370,21 @@ _gst_make_attribute (tree_node attribute_keywords)
        i++, keyword = keyword->v_list.next)
     {
       tree_node value = keyword->v_list.value;
+      OOP result;
       if (value->nodeType != TREE_CONST_EXPR)
 	{
-          OOP result = _gst_execute_statements (NULL, value, UNDECLARED_NONE,
-					        true);
-          value = _gst_make_oop_constant (&value->location, result);
+          result = _gst_execute_statements (NULL, value, UNDECLARED_NONE, true);
           if (!result)
 	    {
 	      _gst_had_error = true;
 	      return _gst_nil_oop;
 	    }
 	}
+      else
+	result = _gst_make_constant_oop (value);
 
       argsArray = OOP_TO_OBJ (argsArrayOOP);
-      argsArray->data[i] = make_constant_oop (value);
+      argsArray->data[i] = result;
     }
 
   messageOOP = _gst_message_new_args (selectorOOP, argsArrayOOP);
@@ -2631,6 +2687,7 @@ _gst_make_new_method (int primitiveIndex,
   inc_ptr incPtr;
 
   maximumStackDepth += numArgs + numTemps;
+  memset (&header, 0, sizeof (method_header));
 
   incPtr = INC_SAVE_POINTER ();
   if (primitiveIndex)
@@ -2809,6 +2866,8 @@ _gst_block_new (int numArgs,
   maximumStackDepth++;		/* just to be sure */
 
   numByteCodes = _gst_bytecode_length (bytecodes);
+
+  memset (&header, 0, sizeof (header));
   header.numArgs = numArgs;
   header.numTemps = numTemps;
   header.depth = maximumStackDepth;
