@@ -17,6 +17,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -47,14 +69,15 @@
 typedef struct gst_file_stream
 {
   OBJ_HEADER;
+  OOP access;
+  OOP fd;
+  OOP file;
+  OOP isPipe;
+  OOP atEnd;
+  OOP peek;
   OOP collection;
   OOP ptr;
   OOP endPtr;
-  OOP access;
-  OOP file;
-  OOP name;
-  OOP isPipe;
-  OOP peek;
   OOP writePtr;
   OOP writeEnd;
 }
@@ -89,6 +112,7 @@ typedef struct gst_namespace
   OOP superspace;
   OOP name;
   OOP subspaces;
+  OOP sharedPools;
   OOP assoc[1];
 
   /* Other, indexable fields that are the associations for this
@@ -108,9 +132,9 @@ typedef struct gst_identity_dictionary
 #define BEHAVIOR_HEADER \
   OBJ_HEADER; \
   OOP		superclass; \
-  OOP		subClasses; \
   OOP		methodDictionary; \
   intptr_t	instanceSpec; \
+  OOP		subClasses; \
   OOP		instanceVariables
 
 typedef struct gst_behavior
@@ -136,12 +160,51 @@ typedef struct gst_behavior
 #define COBJECT_INT_TYPE		FROM_INT(10)
 #define COBJECT_UNSIGNED_INT_TYPE	FROM_INT(11)
 
+/*    3                   2                   1
+    1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
+   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+   |    # fixed fields             |      unused       |I| kind  |1|
+   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+
+   I'm moving it to bits 13-30 (it used to stay at bits 5-30),
+   allocating space for more flags in case they're needed.
+   If you change ISP_NUMFIXEDFIELDS you should modify Behavior.st too.
+   Remember to shift by ISP_NUMFIXEDFIELDS-1 there, since Smalltalk does
+   not see GST_ISP_INTMARK!!  */
+
+enum {
+  ISP_NUMFIXEDFIELDS = 13,
+
+  /* Set if the instances of the class have indexed instance
+     variables.  */
+  ISP_ISINDEXABLE = 32,
+
+  /* These represent the shape of the indexed instance variables of
+     the instances of the class.  */
+  ISP_INDEXEDVARS = 62,
+  ISP_SHAPE = 30,
+
+  /* Set to 1 to mark a SmallInteger.  */
+  GST_ISP_INTMARK = 1
+};
+
 typedef struct gst_class_description
 {
   CLASS_DESCRIPTION_HEADER;
 }
  *gst_class_description;
 
+typedef struct gst_deferred_variable_binding
+{
+  OBJ_HEADER;
+  OOP key;
+  OOP class;
+  OOP defaultDictionary;
+  OOP association;
+  OOP path;
+}
+ *gst_deferred_variable_binding;
+
 typedef struct gst_association
 {
   OBJ_HEADER;
@@ -258,6 +321,7 @@ typedef struct gst_cobject
 {
   OBJ_HEADER;
   OOP type;
+  OOP storage;
 }
  *gst_cobject;
 
@@ -281,8 +345,9 @@ extern OOP _gst_block_closure_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_block_context_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_boolean_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_byte_array_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_byte_stream_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_c_callable_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_c_func_descriptor_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_c_callback_descriptor_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_c_object_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_c_object_type_ctype ATTRIBUTE_HIDDEN;
 extern OOP _gst_c_type_class ATTRIBUTE_HIDDEN;
@@ -296,7 +361,9 @@ extern OOP _gst_compiled_block_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_compiled_code_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_compiled_method_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_context_part_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_continuation_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_date_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_deferred_variable_binding_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_delay_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_dictionary_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_directed_message_class ATTRIBUTE_HIDDEN;
@@ -314,6 +381,7 @@ extern OOP _gst_homed_association_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_identity_dictionary_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_identity_set_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_integer_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_iterable_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_interval_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_large_integer_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_large_negative_integer_class ATTRIBUTE_HIDDEN;
@@ -340,7 +408,6 @@ extern OOP _gst_permission_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_positionable_stream_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_process_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_processor_scheduler_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_random_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_read_stream_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_read_write_stream_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_root_namespace_class ATTRIBUTE_HIDDEN;
@@ -358,12 +425,18 @@ extern OOP _gst_sym_link_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_symbol_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_system_dictionary_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_time_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_token_stream_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_true_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_undefined_object_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_unicode_character_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_unicode_string_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_variable_binding_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_array_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_set_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_key_dictionary_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_value_lookup_table_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_identity_set_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_key_identity_dictionary_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_weak_value_identity_dictionary_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_write_stream_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_processor_oop ATTRIBUTE_HIDDEN;
 
@@ -396,9 +469,16 @@ extern OOP _gst_class_variable_dictionary (OOP class_oop)
   ATTRIBUTE_HIDDEN;
 
 /* This finds the key SYMBOL into the dictionary POOLOOP and, if any,
-   in all of its super-namespaces.  */
-extern OOP _gst_find_shared_pool_variable (OOP poolOOP,
-					   OOP symbol)
+   in all of its super-namespaces.  Returns the association.  */
+extern OOP _gst_namespace_association_at (OOP poolOOP,
+					  OOP symbol)
+  ATTRIBUTE_PURE 
+  ATTRIBUTE_HIDDEN;
+
+/* This finds the key SYMBOL into the dictionary POOLOOP and, if any,
+   in all of its super-namespaces.  Returns the value.  */
+extern OOP _gst_namespace_at (OOP poolOOP,
+			      OOP symbol)
   ATTRIBUTE_PURE 
   ATTRIBUTE_HIDDEN;
 
@@ -444,13 +524,6 @@ extern OOP _gst_message_new_args (OOP selectorOOP,
 				  OOP argsArray) 
   ATTRIBUTE_HIDDEN;
 
-/* Create a new DirectedMessage object for the given RECEIVER,
-   SELECTOR (a Symbol) and Array of arguments.  */
-extern OOP _gst_directed_message_new_args (OOP receiverOOP,
-					   OOP selectorOOP,
-					   OOP argsArray) 
-  ATTRIBUTE_HIDDEN;
-
 /* Returns the name of CLASS_OOP (a Smalltalk Class) */
 extern OOP _gst_get_class_symbol (OOP class_oop)
   ATTRIBUTE_PURE 
@@ -476,21 +549,14 @@ extern OOP _gst_shared_pool_dictionary (OOP class_oop)
   ATTRIBUTE_PURE 
   ATTRIBUTE_HIDDEN;
 
-/* Creates a new CObject pointing to cObjPtr, extracting the name of
-   the class to be instantiated from the CType, TYPEOOP.  */
-extern OOP _gst_c_object_new_typed (PTR cObjPtr,
-				    OOP typeOOP) 
-  ATTRIBUTE_HIDDEN;
-
-/* Allocates a new CObject by malloc-ing SIZE bytes; CLASS_OOP is the
-   class to be instantiated.  */
-extern OOP _gst_alloc_cobject (OOP class_oop,
-			       size_t size) 
-  ATTRIBUTE_HIDDEN;
-
-/* Creates a new CType that when passed to _gst_c_object_new_typed
-   creates an instance of COBJECTSUBCLASSOOP.  */
-extern OOP _gst_c_type_new (OOP cObjectSubclassOOP) 
+/* Creates a new CObject pointing to cObjOfs bytes in BASEOOP (or
+   at the absolute address cObjOfs if BASEOOP is NULL), extracting the class
+   to be instantiated from the CType, TYPEOOP, or using the provided
+   class if TYPEOOP is nil.  */
+extern OOP _gst_c_object_new_base (OOP baseOOP,
+			           uintptr_t cObjOfs,
+			           OOP typeOOP,
+			           OOP defaultClassOOP) 
   ATTRIBUTE_HIDDEN;
 
 /* Creates a new String with LEN indexed instance variables.  */
@@ -588,10 +654,10 @@ extern void _gst_set_oop_bytes (OOP byteArrayOOP,
 extern void _gst_free_cobject (OOP cObjOOP) 
   ATTRIBUTE_HIDDEN;
 
-/* Loads the contents of the global variable from the Smalltalk
-   dictionary after NUMOOPS objects have been loaded from an
-   image.  */
-extern mst_Boolean _gst_init_dictionary_on_image_load (size_t numOOPs) 
+/* Loads the contents of the global variables from the Smalltalk dictionary
+   after an image has been restored.  PRIM_TABLE_MATCHES if true if the
+   table of primitives is already set up correctly.  */
+extern mst_Boolean _gst_init_dictionary_on_image_load (mst_Boolean prim_table_matches) 
   ATTRIBUTE_HIDDEN;
 
 /* Transforms a primitive name into a primitive index, looking up
