@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2006
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2006,2007
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -90,9 +90,6 @@ const char *_gst_image_file_path = NULL;
    home directory.  */
 const char *_gst_user_file_base_path = NULL;
 
-/* The path to the executable, derived from argv[0].  */
-const char *_gst_executable_path = NULL;
-
 /* Whether to look for user files.  */
 static mst_Boolean no_user_files = false;
 
@@ -170,6 +167,11 @@ static const char standard_files[] = {
   "False.st\0"
   "True.st\0"
   "Magnitude.st\0"
+  "LookupKey.st\0"
+  "DeferBinding.st\0"
+  "Association.st\0"
+  "HomedAssoc.st\0"
+  "VarBinding.st\0"
   "Integer.st\0"
   "Date.st\0"
   "Time.st\0"
@@ -183,21 +185,18 @@ static const char standard_files[] = {
   "SmallInt.st\0"
   "Character.st\0"
   "UniChar.st\0"
-  "LookupKey.st\0"
-  "Association.st\0"
-  "HomedAssoc.st\0"
-  "VarBinding.st\0"
   "Link.st\0"
   "Process.st\0"
   "CallinProcess.st\0"
-  "CompildCode.st\0"
-  "CompildMeth.st\0"
-  "CompiledBlk.st\0"
+  "Iterable.st\0"
   "Collection.st\0"
   "SeqCollect.st\0"
   "LinkedList.st\0"
   "Semaphore.st\0"
   "ArrayColl.st\0"
+  "CompildCode.st\0"
+  "CompildMeth.st\0"
+  "CompiledBlk.st\0"
   "Array.st\0"
   "ByteArray.st\0"
   "CharArray.st\0"
@@ -225,19 +224,18 @@ static const char standard_files[] = {
   "ReadStream.st\0"
   "WriteStream.st\0"
   "RWStream.st\0"
-  "ByteStream.st\0"
   "UndefObject.st\0"
   "ProcSched.st\0"
   "Delay.st\0"
   "SharedQueue.st\0"
-  "Behavior.st\0"
-  "ClassDesc.st\0"
-  "Class.st\0"
-  "Metaclass.st\0"
   "ContextPart.st\0"
   "MthContext.st\0"
   "BlkContext.st\0"
   "BlkClosure.st\0"
+  "Behavior.st\0"
+  "ClassDesc.st\0"
+  "Class.st\0"
+  "Metaclass.st\0"
   "Continuation.st\0"
   "Memory.st\0"
   "MethodInfo.st\0"
@@ -263,17 +261,22 @@ static const char standard_files[] = {
   /* C call-out facilities */
   "CObject.st\0"
   "CType.st\0"
+  "CCallable.st\0"
   "CFuncs.st\0"
+  "CCallback.st\0"
   "CStruct.st\0"
 
-  /* Exception handling */
+  /* Exception handling and ProcessEnvironment */
+  "ProcEnv.st\0"
   "ExcHandling.st\0"
   "AnsiExcept.st\0"
 
   /* Virtual filesystem layer */
+  "FilePath.st\0"
   "File.st\0"
   "Directory.st\0"
   "VFS.st\0"
+  "VFSZip.st\0"
   "URL.st\0"
   "FileStream.st\0"
 
@@ -306,7 +309,6 @@ _gst_smalltalk_args (int argc,
 {
   smalltalk_argc = argc;
   smalltalk_argv = argv;
-  _gst_executable_path = _gst_find_executable (argv[0]);
 }
 
    
@@ -328,6 +330,9 @@ _gst_initialize (const char *kernel_dir,
   _gst_smalltalk_initialized = true;
   _gst_init_snprintfv ();
 
+  if (!_gst_executable_path)
+    _gst_executable_path = DEFAULT_EXECUTABLE;
+
   /* By default, apply this kludge fpr OSes such as Windows and MS-DOS
      which have no concept of home directories.  */
   if (home == NULL)
@@ -350,6 +355,8 @@ _gst_initialize (const char *kernel_dir,
 	}
     }
 
+  /* For the image file, it is okay to find none if we can/should rebuild
+     the image file.  */
   if (image_file
       && (flags & (GST_REBUILD_IMAGE | GST_MAYBE_REBUILD_IMAGE)) == 0
       && !_gst_file_is_readable (image_file))
@@ -363,6 +370,8 @@ _gst_initialize (const char *kernel_dir,
 	}
     }
 
+  /* The image path can be used as the default kernel path, so we split
+     it anyway into directory+filename.  */
   if (image_file)
     {
       const char *p;
@@ -393,26 +402,31 @@ _gst_initialize (const char *kernel_dir,
     }
   else
     {
-      /* No image file given, we use the system default but revert to the
+      /* No image file given, we use the system default or revert to the
 	 current directory.  */
-      if (_gst_file_is_readable (IMAGE_PATH))
-        _gst_image_file_path = IMAGE_PATH;
+      str = _gst_relocate_path (IMAGE_PATH);
+      if (_gst_file_is_readable (str))
+        _gst_image_file_path = str;
       else
-        _gst_image_file_path = xstrdup (currentDirectory);
+	{
+          free (str);
+          _gst_image_file_path = xstrdup (currentDirectory);
+	}
+
       flags |= GST_IGNORE_BAD_IMAGE_PATH;
       image_file = "gst.im";
     }
 
   if (!kernel_dir)
     {
-      if (_gst_file_is_readable (KERNEL_PATH))
-        kernel_dir = KERNEL_PATH;
-      else
+      str = _gst_relocate_path (KERNEL_PATH);
+      if (!_gst_file_is_readable (str))
 	{
-          char *kernel_file_path;
-	  asprintf (&kernel_file_path, "%s/kernel", _gst_image_file_path);
-	  kernel_dir = kernel_file_path;
+          free (str);
+	  asprintf (&str, "%s/kernel", _gst_image_file_path);
 	}
+
+      kernel_dir = str;
     }
 
   xfree (currentDirectory);
@@ -526,29 +540,27 @@ _gst_initialize (const char *kernel_dir,
 mst_Boolean
 ok_to_load_binary (void)
 {
-  time_t imageFileTime;
   const char *fileName;
 
-  imageFileTime = _gst_get_file_modify_time (_gst_binary_image_name);
-
-  if (imageFileTime == 0)	/* not found */
+  if (!_gst_file_is_readable (_gst_binary_image_name))
     return (false);
 
   for (fileName = standard_files; *fileName; fileName += strlen (fileName) + 1)
     {
       char *fullFileName = _gst_find_file (fileName, GST_DIR_KERNEL);
-      mst_Boolean ok = (imageFileTime > _gst_get_file_modify_time (fullFileName));
+      mst_Boolean ok = _gst_file_is_newer (_gst_binary_image_name,
+					   fullFileName);
       xfree (fullFileName);
       if (!ok)
         return (false);
     }
 
   if (site_pre_image_file
-      && imageFileTime <= _gst_get_file_modify_time (site_pre_image_file))
+      && !_gst_file_is_newer (_gst_binary_image_name, site_pre_image_file))
     return (false);
 
   if (user_pre_image_file
-      && imageFileTime <= _gst_get_file_modify_time (user_pre_image_file))
+      && !_gst_file_is_newer (_gst_binary_image_name, user_pre_image_file))
     return (false);
 
   return (true);
@@ -563,12 +575,15 @@ load_standard_files (void)
     {
       if (!_gst_process_file (fileName, GST_DIR_KERNEL))
 	{
-	  _gst_errorf ("can't find system file '%s'", fileName);
+	  _gst_errorf ("couldn't load system file '%s': %s", fileName,
+		       strerror (errno));
 	  _gst_errorf ("image bootstrap failed, use option --kernel-directory");
 	  return 1;
 	}
     }
 
+  _gst_msg_sendf (NULL, "%v %o relocate", _gst_file_segment_class);
+
   if (site_pre_image_file)
     _gst_process_file (site_pre_image_file, GST_DIR_ABS);
 
@@ -599,11 +614,7 @@ _gst_find_file (const char *fileName,
 		dir == GST_DIR_BASE ? "" : LOCAL_KERNEL_DIR_NAME "/",
 		fileName);
 
-      if (_gst_file_is_readable (localFileName)
-          /* If the system file is newer, use the system file instead.  */
-          && (!_gst_file_is_readable (fullFileName) ||
-	      _gst_get_file_modify_time (localFileName) >=
-	      _gst_get_file_modify_time (fullFileName)))
+      if (_gst_file_is_newer (localFileName, fullFileName))
 	{
 	  xfree (fullFileName);
 	  return localFileName;
