@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -25,7 +47,7 @@
  * 
  * You should have received a copy of the GNU General Public License along with
  * GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
  *
  ***********************************************************************/
 
@@ -33,6 +55,61 @@
 #define GST_GSTPRIV_H
 
 #include "config.h"
+
+#include <sys/types.h>
+#include <stdio.h>
+#include <stddef.h>
+#include <setjmp.h>
+#include <assert.h>
+#include <stdlib.h>
+#include <string.h>
+#include <obstack.h>
+#include <fcntl.h>
+#include <stdarg.h>
+#include <math.h>
+#include <float.h>
+#include <errno.h>
+#include <sys/stat.h>
+#include <limits.h>
+#include <poll.h>
+#include <ctype.h>
+#include <wchar.h>
+#include <dirent.h>
+#include <sys/time.h>
+#include <time.h>
+
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
+
+#ifdef HAVE_EXECINFO_H
+#include <execinfo.h>
+#endif
+
+#ifdef HAVE_SYS_FILE_H
+#include <sys/file.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#ifdef HAVE_SYS_MMAN_H
+#include <sys/mman.h>
+#endif
+
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
+
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#endif
+
+#ifdef HAVE_SIGSEGV_H
+#include "sigsegv.h"
+#endif
+
 #include "gst.h"
 
 /* Convenience macros to test the versions of GCC.  Note - they won't
@@ -48,7 +125,7 @@
 /* For internal functions, we can use the ELF hidden attribute to
    improve code generation.  Unluckily, this is only in GCC 3.2 and
    later */
-#if GNUC_PREREQ (3, 3)
+#ifdef HAVE_VISIBILITY_HIDDEN
 #define ATTRIBUTE_HIDDEN __attribute__ ((visibility ("hidden")))
 #else
 #define ATTRIBUTE_HIDDEN
@@ -121,6 +198,11 @@
 #define PREFETCH_ADDR(x, k) \
   DO_PREFETCH ((x), 0, (k));
 
+/* Synchronization primitives.  */
+#if !defined HAVE_SYNC_BUILTINS && defined __GNUC__
+#define __sync_synchronize()	__asm__ ("" : : : "memory")
+#endif
+
 /* Kill a warning when using GNU C.  Note that this allows using
    break or continue inside a macro, unlike do...while(0) */
 #ifdef __GNUC__
@@ -136,9 +218,9 @@
    special marking of untrusted objects.  Note that the code in the
    class library to perform the security checks will be present
    notwithstanding the setting of this flag, but they will be disabled
-   because the corresponding primitives will be made non-working.
-   It is undefined because the Makefiles take care of defining it for
-   security-enabled builds.  */
+   because the corresponding primitives will be made non-working.  We
+   define it here with no configure-time options because it causes
+   testsuite failures.  */
 #define ENABLE_SECURITY
 
 /* OPTIMIZE disables many checks, including consistency checks at GC
@@ -234,35 +316,36 @@ enum {
 
   /* The grouping of all the flags which are not valid across image
      saves and loads.  */
-  F_RUNTIME = 0xFFF000U,
-
-  /* Set if the OOP is currently unused.  */
-  F_FREE = 0x10U,
+  F_RUNTIME = 0xFF8000U,
 
   /* Set if the references to the instance variables of the object
      are weak.  */
-  F_WEAK = 0x20U,
+  F_WEAK = 0x10U,
 
   /* Set if the object is read-only.  */
-  F_READONLY = 0x40U,
+  F_READONLY = 0x20U,
 
   /* Set if the object is a context and hence its instance variables
      are only valid up to the stack pointer.  */
-  F_CONTEXT = 0x80U,
+  F_CONTEXT = 0x40U,
 
   /* Answer whether we want to mark the key based on references found
      outside the object.  */
-  F_EPHEMERON = 0x100U,
+  F_EPHEMERON = 0x80U,
 
   /* Set for objects that live in oldspace.  */
-  F_OLD = 0x200U,
+  F_OLD = 0x100U,
 
   /* Set together with F_OLD for objects that live in fixedspace.  */
-  F_FIXED = 0x400U,
+  F_FIXED = 0x200U,
 
   /* Set for untrusted classes, instances of untrusted classes,
      and contexts whose receiver is untrusted.  */
-  F_UNTRUSTED = 0x800U,
+  F_UNTRUSTED = 0x400U,
+
+  /* Set for objects that were loaded from the image.  We never
+     garbage collect their contents, only the OOPs.  */
+  F_LOADED = 0x800U,
 
   /* Set to the number of bytes unused in an object with byte-sized
      instance variables.  Note that this field and the following one
@@ -305,8 +388,8 @@ enum {
    ((oop)->flags |= (untr) ? F_UNTRUSTED : 0))
 
 #else
-#define IS_OOP_UNTRUSTED(oop) 0
-#define MAKE_OOP_UNTRUSTED(oop, untr) 0
+#define IS_OOP_UNTRUSTED(oop) (false)
+#define MAKE_OOP_UNTRUSTED(oop, untr) ((void)0)
 #endif
 
 /* Set whether an object, OOP, has ephemeron semantics.  */
@@ -314,66 +397,6 @@ enum {
   (oop)->flags |= F_EPHEMERON;
 
 
-/*
-      3                   2                   1 
-    1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
-   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-   |    # fixed fields             |      unused       |I| kind  |1|
-   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-
-   I'm moving it to bits 13-30 (it used to stay at bits 5-30),
-   allocating space for more flags in case they're needed.
-   If you change ISP_NUMFIXEDFIELDS you should modify Behavior.st too.
-   Remember to shift by ISP_NUMFIXEDFIELDS-1 there, since Smalltalk does
-   not see ISP_INTMARK!!
-
-   Keep these in sync with _gst_sizes, in dict.c.
-
-   FIXME: these should be exported in a pool dictionary.  */
-
-enum {
-  /* This is a shift count.  */
-  ISP_NUMFIXEDFIELDS = 13,
-
-  /* Set if the indexed instance variables are pointers.  */
-  ISP_FIXED = 0,
-  ISP_SCHAR = 32,
-  ISP_UCHAR = 34,
-  ISP_SHORT = 36,
-  ISP_USHORT = 38,
-  ISP_INT = 40,
-  ISP_UINT = 42,
-  ISP_FLOAT = 44,
-  ISP_INT64 = 46,
-  ISP_UINT64 = 48,
-  ISP_DOUBLE = 50,
-  ISP_LAST_SCALAR = 50,
-  ISP_POINTER = 62,
-
-#if SIZEOF_LONG == 8
-  ISP_LONG = ISP_INT64,
-  ISP_ULONG = ISP_UINT64,
-  ISP_LAST_UNALIGNED = ISP_FLOAT,
-#else
-  ISP_LONG = ISP_INT,
-  ISP_ULONG = ISP_UINT,
-  ISP_LAST_UNALIGNED = ISP_USHORT,
-#endif
-
-  /* Set if the instances of the class have indexed instance
-     variables.  */
-  ISP_ISINDEXABLE = 32,
-
-  /* These represent the shape of the indexed instance variables of
-     the instances of the class.  */
-  ISP_INDEXEDVARS = 62,
-  ISP_SHAPE = 30,
-
-  /* Set to 1 to mark a SmallInteger.  */
-  ISP_INTMARK = 1
-};
-
-
 /* the current execution stack pointer */
 #ifndef ENABLE_JIT_TRANSLATION
 # define sp		_gst_sp
@@ -504,7 +527,7 @@ extern OOP _gst_nil_oop
 #define ST_INT_SIZE        ((sizeof (PTR) * 8) - 2)
 #define MAX_ST_INT         ((1L << ST_INT_SIZE) - 1)
 #define MIN_ST_INT         ( ~MAX_ST_INT)
-#define INT_OVERFLOW(i)    ( (i) > MAX_ST_INT || (i) < MIN_ST_INT )
+#define INT_OVERFLOW(i)    (((i) ^ ((i) << 1)) < 0)
 #define OVERFLOWING_INT    (MAX_ST_INT + 1)
 
 #define INCR_INT(i)         ((OOP) (((intptr_t)i) + 2))	/* 1 << 1 */
@@ -514,19 +537,24 @@ extern OOP _gst_nil_oop
    the correct job (that is, on 32-bit little-endian systems)
    because they are likely to be optimized.  */
 
-#if SIZEOF_INTPTR_T == 4
-# if defined(WORDS_BIGENDIAN) || !defined (HAVE_INET_SOCKETS)
+#if SIZEOF_OOP == 4
+# if !defined(WORDS_BIGENDIAN) && defined (HAVE_SOCKETS)
+#  define BYTE_INVERT(x) htonl((x))
+# elif defined _OS_OSBYTEORDERPPC_H
+#  define BYTE_INVERT(x) OSReadSwapInt32(&(x), 0)
+# else
 #  define BYTE_INVERT(x) \
         ((uintptr_t)((((uintptr_t)(x) & 0x000000ffU) << 24) | \
                      (((uintptr_t)(x) & 0x0000ff00U) <<  8) | \
                      (((uintptr_t)(x) & 0x00ff0000U) >>  8) | \
                      (((uintptr_t)(x) & 0xff000000U) >> 24)))
-# else
-#  define BYTE_INVERT(x) htonl((x))
 # endif
 
-#else /* SIZEOF_INTPTR_T == 8 */
-# define BYTE_INVERT(x) \
+#else /* SIZEOF_OOP == 8 */
+# if defined _OS_OSBYTEORDERPPC_H
+#  define BYTE_INVERT(x) OSReadSwapInt64(&(x), 0)
+# else
+#  define BYTE_INVERT(x) \
         ((uintptr_t)((((uintptr_t)(x) & 0x00000000000000ffU) << 56) | \
                      (((uintptr_t)(x) & 0x000000000000ff00U) << 40) | \
                      (((uintptr_t)(x) & 0x0000000000ff0000U) << 24) | \
@@ -535,7 +563,8 @@ extern OOP _gst_nil_oop
                      (((uintptr_t)(x) & 0x0000ff0000000000U) >> 24) | \
                      (((uintptr_t)(x) & 0x00ff000000000000U) >> 40) | \
                      (((uintptr_t)(x) & 0xff00000000000000U) >> 56)))
-#endif /* SIZEOF_INTPTR_T == 8 */
+# endif
+#endif /* SIZEOF_OOP == 8 */
 
 /* The standard min/max macros...  */
 
@@ -549,69 +578,6 @@ extern OOP _gst_nil_oop
 #define MIN(x, y) 		( ((x) > (y)) ? (y) : (x) )
 #endif
 
-#include <sys/types.h>
-#include <stdio.h>
-#include <stddef.h>
-#include <setjmp.h>
-#include <assert.h>
-#include <stdlib.h>
-#include <string.h>
-#include <obstack.h>
-#include <fcntl.h>
-#include <stdarg.h>
-#include <math.h>
-#include <float.h>
-#include <errno.h>
-#include <sys/stat.h>
-#include <limits.h>
-#include <poll.h>
-#include <ctype.h>
-
-#ifdef HAVE_DIRENT_H
-# include <dirent.h>
-# define NAMLEN(dirent) strlen((dirent)->d_name)
-#else
-# define dirent direct
-# define NAMLEN(dirent) (dirent)->d_namlen
-# ifdef HAVE_SYS_NDIR_H
-#  include <sys/ndir.h>
-# endif
-# ifdef HAVE_SYS_DIR_H
-#  include <sys/dir.h>
-# endif
-# ifdef HAVE_NDIR_H
-#  include <ndir.h>
-# endif
-#endif
-
-#if defined(HAVE_SYS_TIME_H) || defined(TIME_WITH_SYS_TIME)
-# include <sys/time.h>
-#endif
-#if !defined(HAVE_SYS_TIME_H) || defined(TIME_WITH_SYS_TIME)
-# include <time.h>
-#endif
-
-#ifdef HAVE_SYS_RESOURCE_H
-#include <sys/resource.h>
-#endif
-
-#ifdef HAVE_EXECINFO_H
-#include <execinfo.h>
-#endif
-
-#ifdef HAVE_SYS_FILE_H
-#include <sys/file.h>
-#endif
-
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-
-#ifdef HAVE_SYS_MMAN_H
-#include <sys/mman.h>
-#endif
-
-#include "stdintx.h"
 #include "ansidecl.h"
 #include "signalx.h"
 #include "mathl.h"
@@ -624,34 +590,45 @@ extern OOP _gst_nil_oop
 #include "rbtrees.h"
 
 #include "tree.h"
+#include "files.h"
 #include "input.h"
 #include "callin.h"
 #include "cint.h"
 #include "dict.h"
 #include "events.h"
-#include "gstpub.h"
 #include "heap.h"
 #include "lex.h"
-#include "lib.h"
+#include "gst-parse.h"
 #include "oop.h"
 #include "byte.h"
+#include "sym.h"
 #include "comp.h"
 #include "interp.h"
 #include "opt.h"
 #include "save.h"
 #include "str.h"
 #include "sysdep.h"
-#include "sym.h"
 #include "xlat.h"
 #include "mpz.h"
 #include "print.h"
 #include "security.h"
-#include "match.h"
 
 /* Include this last, it has the bad habit of #defining printf
    and this fools gcc's __attribute__ (format) */
 #include "snprintfv/printf.h"
 
+#undef obstack_init
+#define obstack_init(h)                                         \
+  _obstack_begin ((h), 0, ALIGNOF_LONG_DOUBLE,                \
+                  (void *(*) (long)) obstack_chunk_alloc,       \
+                  (void (*) (void *)) obstack_chunk_free)
+
+#undef obstack_begin
+#define obstack_begin(h, size)                                  \
+  _obstack_begin ((h), (size), ALIGNOF_LONG_DOUBLE,           \
+                  (void *(*) (long)) obstack_chunk_alloc,       \
+                  (void (*) (void *)) obstack_chunk_free)
+
 #include "oop.inl"
 #include "dict.inl"
 #include "interp.inl"
