@@ -10,7 +10,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -55,6 +55,7 @@
  ***********************************************************************/
 
 #include "gstpriv.h"
+#include "lock.h"
 
 
 
@@ -114,24 +115,21 @@
    to speed up these messages for Arrays, Strings, and ByteArrays.  */
 #define METHOD_CACHE_SIZE		(1 << 11)
 
-/* Length of the queue of Semaphores to be signaled at the next
-   sequence point.  */
-#define ASYNC_QUEUE_SIZE		100
-
-
 typedef struct async_queue_entry
 {
-  OOP sem;
-  mst_Boolean unregister;
+  void (*func) (OOP);
+  OOP data;
 }
 async_queue_entry;
 
 typedef struct interp_jmp_buf
 {
-  struct interp_jmp_buf *next;
-  int suspended;
-  OOP processOOP;
   jmp_buf jmpBuf;
+  struct interp_jmp_buf *next;
+  unsigned short suspended;
+  unsigned char interpreter;
+  unsigned char interrupted;
+  OOP processOOP;
 }
 interp_jmp_buf;
 
@@ -218,9 +216,6 @@ OOP _gst_this_method = NULL;
 /* Signal this semaphore at the following instruction.  */
 static OOP single_step_semaphore = NULL;
 
-/* Answer whether we are in the interpreter or in application code.  */
-static mst_Boolean in_interpreter = false;
-
 /* CompiledMethod cache which memoizes the methods and some more
    information for each class->selector pairs.  */
 static method_cache_entry method_cache[METHOD_CACHE_SIZE] CACHELINE_ALIGNED;
@@ -248,26 +243,23 @@ static int class_cache_prim;
 #endif
 
 /* Queue for async (outside the interpreter) semaphore signals */
+gl_lock_define (static, async_queue_lock);
+static mst_Boolean async_queue_enabled = true;
 static int async_queue_index;
-static async_queue_entry queued_async_signals[ASYNC_QUEUE_SIZE]
-  CACHELINE_ALIGNED;
+static int async_queue_size;
+static async_queue_entry *queued_async_signals;
+
+static int async_queue_index_sig;
+static async_queue_entry queued_async_signals_sig[NSIG];
 
 /* When not NULL, this causes the byte code interpreter to immediately
    send the message whose selector is here to the current stack
    top.  */
 const char *_gst_abort_execution = NULL;
 
-/* Set to true when some special action must be done at the next
-   sequence point.  */
-volatile mst_Boolean _gst_except_flag = false;
-
 /* Set to non-nil if a process must preempt the current one.  */
 static OOP switch_to_process;
 
-/* Set to true if the currently running process has disabled
-   external signals.  */
-static mst_Boolean disable_preemption;
-
 /* Set to true if it is time to switch process in a round-robin
    time-sharing fashion.  */
 static mst_Boolean time_to_preempt;
@@ -328,17 +320,12 @@ static OOP next_scheduled_process (void);
    CONTEXTOOP context, and answer it.  */
 static OOP create_callin_process (OOP contextOOP);
 
-/* Sets flags so that the interpreter starts returning immediately from
-   whatever byte codes it's executing.  It returns via a normal message
-   send of the unary selector MSG, so that the world is in a consistent 
-   state when it's done.  */
-static void stop_executing (const char *msg);
-
 /* Set a timer at the end of which we'll preempt the current process.  */
 static void set_preemption_timer (void);
 
-/* Same as _gst_parse_stream, but creating a reentrancy_jmpbuf.  */
-static void parse_stream_with_protection (mst_Boolean method);
+/* Same as _gst_parse_stream, but creating a reentrancy_jmpbuf.  Returns
+   true if interrupted. */
+static mst_Boolean parse_stream_with_protection (mst_Boolean method);
 
 /* Put the given process to sleep by rotating the list of processes for
    PROCESSOOP's priority (i.e. it was the head of the list and becomes
@@ -370,11 +357,6 @@ static void mark_semaphore_oops (void);
    switch to at the next sequence point).  */
 static void copy_semaphore_oops (void);
 
-/* Signal the given SEMAPHOREOOP and if processes were queued on it
-   resume the one that has waited for the longest time and is still
-   alive.  */
-static void sync_signal (OOP semaphoreOOP);
-
 /* Suspend execution of PROCESSOOP.  */
 static void suspend_process (OOP processOOP);
 
@@ -424,9 +406,10 @@ static inline OOP create_args_array (int numArgs);
    the arguments in the block context, which have been copied out of
    the caller's context. 
 
-   On failure return true, on success (i.e. if NUMARGS matches what
-   the BlockClosure says) return false.  */
-static mst_Boolean send_block_value (int numArgs);
+   The block should accept between NUMARGS - CULL_UP_TO and
+   NUMARGS arguments.  If this is not true (failure) return true;
+   on success return false.  */
+static mst_Boolean send_block_value (int numArgs, int cull_up_to);
 
 /* This is a kind of simplified _gst_send_message_internal that,
    instead of setting up a context for a particular receiver, stores
@@ -516,54 +499,28 @@ static mst_Boolean unwind_to (OOP returnContextOOP);
    doing a local return.  */
 static mst_Boolean disable_non_unwind_contexts (OOP returnContextOOP);
 
-/* Called to handle signals that are not passed to the Smalltalk
-   program, such as interrupts or segmentation violation.  In the
-   latter case, try to show a method invocation backtrace if possibly,
-   otherwise try to show where the system was in the file it was
-   processing when the error occurred.  */
-static RETSIGTYPE interrupt_handler (int sig);
-
 /* Called to preempt the current process after a specified amount
    of time has been spent in the GNU Smalltalk interpreter.  */
 #ifdef ENABLE_PREEMPTION
 static RETSIGTYPE preempt_smalltalk_process (int sig);
 #endif
 
-/* This macro acts as a block statement (if, for, while); it
-   accepts a pointer to an interp_jmp_buf and executes its body
-   so that the current process is suspended and SIGINT breaks
-   out of it.  */
-#define PROTECT_CURRENT_PROCESS_WITH(jb) \
-  for ((jb)->next = reentrancy_jmp_buf, \
-       (jb)->suspended = 0, \
-       (jb)->processOOP = get_active_process (), \
-       _gst_register_oop ((jb)->processOOP), \
-       reentrancy_jmp_buf = (jb), \
-       in_interpreter = false; \
-       !in_interpreter; \
-       in_interpreter = true, \
-       _gst_unregister_oop ((jb)->processOOP), \
-       reentrancy_jmp_buf = reentrancy_jmp_buf->next) \
-    if (setjmp (reentrancy_jmp_buf->jmpBuf) != 0) \
-      continue; \
-    else \
+/* Push an execution state for process PROCESSOOP.  The process is
+   used for two reasons: 1) it is suspended if there is a call-in
+   while the execution state is on the top of the stack; 2) it is
+   sent #userInterrupt if the user presses Ctrl-C.  */
+static void push_jmp_buf (interp_jmp_buf *jb,
+			  int for_interpreter,
+			  OOP processOOP);
 
-/* This macro acts as a block statement (if, for, while); it
-   accepts a pointer to an interp_jmp_buf and executes its body
-   so that the current process is not suspended (like in
-   asynchronous C call-outs) but SIGINT breaks out of it.  */
-#define PROTECT_FROM_INTERRUPT_WITH(jb) \
-  for ((jb)->next = reentrancy_jmp_buf, \
-       (jb)->suspended = 0, \
-       (jb)->processOOP = _gst_nil_oop, \
-       reentrancy_jmp_buf = (jb), \
-       in_interpreter = false; \
-       !in_interpreter; \
-       in_interpreter = true, \
-       reentrancy_jmp_buf = reentrancy_jmp_buf->next) \
-    if (setjmp (reentrancy_jmp_buf->jmpBuf) != 0) \
-      continue; \
-    else \
+/* Pop an execution state.  Return true if the interruption has to
+   be propagated up.  */
+static mst_Boolean pop_jmp_buf (void);
+
+/* Jump out of the top execution state.  This is used by C call-out
+   primitives to jump out repeatedly until a Smalltalk process is
+   encountered and terminated.  */
+static void stop_execution (void);
 
 /* Pick a process that is the highest-priority process different from
    the currently executing one, and schedule it for execution after
@@ -579,16 +536,19 @@ static RETSIGTYPE preempt_smalltalk_process (int sig);
 /* Tell the interpreter that special actions are needed as soon as a
    sequence point is reached.  */
 #ifdef ENABLE_JIT_TRANSLATION
-#define SET_EXCEPT_FLAG(x) _gst_except_flag = (x);
+mst_Boolean _gst_except_flag = false;
+#define SET_EXCEPT_FLAG(x) \
+  do { _gst_except_flag = (x); __sync_synchronize (); } while (0)
+
 #else
 static void * const *global_monitored_bytecodes;
 static void * const *global_normal_bytecodes;
 static void * const *dispatch_vec;
 
-#define SET_EXCEPT_FLAG(x) do {						\
-  _gst_except_flag = (x);						\
+#define SET_EXCEPT_FLAG(x) do { \
   dispatch_vec = (x) ? global_monitored_bytecodes : global_normal_bytecodes; \
-} while(0)
+  __sync_synchronize (); \
+} while (0)
 #endif
 
 /* Answer an hash value for a send of the SENDSELECTOR message, when
@@ -1322,7 +1282,7 @@ change_process_context (OOP newProcess)
   OOP processOOP;
   gst_process process;
   gst_processor_scheduler processor;
-  mst_Boolean preemption_was_disabled;
+  mst_Boolean enable_async_queue;
 
   switch_to_process = _gst_nil_oop;
 
@@ -1339,15 +1299,14 @@ change_process_context (OOP newProcess)
   if (processOOP != newProcess)
     {
       process = (gst_process) OOP_TO_OBJ (processOOP);
-      preemption_was_disabled = disable_preemption;
 
       if (!IS_NIL (processOOP) && !is_process_terminating (processOOP))
         process->suspendedContext = _gst_this_context_oop;
 
       processor->activeProcess = newProcess;
       process = (gst_process) OOP_TO_OBJ (newProcess);
-      disable_preemption = !IS_NIL (process->interrupts) &&
-		           TO_INT (process->interrupts) < 0;
+      enable_async_queue = IS_NIL (process->interrupts)
+		           || TO_INT (process->interrupts) >= 0;
 
       resume_suspended_context (process->suspendedContext);
 
@@ -1366,13 +1325,7 @@ change_process_context (OOP newProcess)
            section has terminated (see RecursionLock>>#critical: for an
            example).  */
 
-      if (preemption_was_disabled ^ disable_preemption)
-	{
-          if (disable_preemption)
-            _gst_disable_interrupts();
-          else
-            _gst_enable_interrupts();
-	}
+      async_queue_enabled = enable_async_queue;
     }
 }
 
@@ -1462,8 +1415,8 @@ add_first_link (OOP semaphoreOOP,
     sem->lastLink = processOOP;
 }
 
-void
-suspend_process (OOP processOOP)
+static void
+remove_process_from_list (OOP processOOP)
 {
   gst_semaphore sem;
   gst_process process, lastProcess;
@@ -1504,6 +1457,12 @@ suspend_process (OOP processOOP)
     }
 
   process->nextLink = _gst_nil_oop;
+}
+
+void
+suspend_process (OOP processOOP)
+{
+  remove_process_from_list (processOOP);
   if (get_scheduled_process() == processOOP)
     ACTIVE_PROCESS_YIELD ();
 }
@@ -1579,50 +1538,91 @@ is_empty (OOP processListOOP)
 }
 
 
-void
-sync_signal (OOP semaphoreOOP)
+mst_Boolean
+_gst_sync_signal (OOP semaphoreOOP, mst_Boolean incr_if_empty)
 {
   gst_semaphore sem;
   OOP freedOOP;
 
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
-  do
+  for (;;)
     {
       /* printf ("signal %O %O\n", semaphoreOOP, sem->firstLink); */
       if (is_empty (semaphoreOOP))
 	{
-	  sem->signals = INCR_INT (sem->signals);
-	  break;
+	  if (incr_if_empty)
+	    sem->signals = INCR_INT (sem->signals);
+	  return false;
 	}
+
       freedOOP = remove_first_link (semaphoreOOP);
 
       /* If they terminated this process, well, try another */
+      if (resume_process (freedOOP, false))
+	return true;
     }
-  while (!resume_process (freedOOP, false));
+}
+
+static void
+do_async_signal (OOP semaphoreOOP)
+{
+  _gst_sync_signal (semaphoreOOP, true);
+}
+
+void
+_gst_async_call (void (*func) (OOP), OOP arg)
+{
+  if (_gst_signal_count)
+    {
+      /* Async-signal-safe version.  Because the discriminant is
+	 _gst_signal_count, can only be called from within libgst.la
+	 (in particular events.c).  Also, because of that we assume
+	 interrupts are already disabled.  */
+
+      if (async_queue_index_sig == NSIG)
+        {
+	  static const char errmsg[] = "Asynchronous signal queue full!\n";
+	  write (2, errmsg, sizeof (errmsg) - 1);
+	  abort ();
+	}
+
+      queued_async_signals_sig[async_queue_index_sig].func = func;
+      queued_async_signals_sig[async_queue_index_sig++].data = arg;
+    }
+  else
+    {
+      /* Thread-safe version for the masses.  */
+
+      gl_lock_lock (async_queue_lock);
+      if (async_queue_index == async_queue_size)
+        {
+          async_queue_size *= 2;
+          queued_async_signals =
+	    realloc (queued_async_signals,
+		 sizeof (async_queue_entry) * async_queue_size);
+	}
+
+      queued_async_signals[async_queue_index].func = func;
+      queued_async_signals[async_queue_index++].data = arg;
+      gl_lock_unlock (async_queue_lock);
+    }
+  
+  SET_EXCEPT_FLAG (true);
 }
 
 void
 _gst_async_signal (OOP semaphoreOOP)
 {
-  _gst_disable_interrupts ();	/* block out everything! */
-  
-  queued_async_signals[async_queue_index].sem = semaphoreOOP;
-  queued_async_signals[async_queue_index++].unregister = false;
-  
-  SET_EXCEPT_FLAG (true);
-  _gst_enable_interrupts ();
+  _gst_async_call (do_async_signal, semaphoreOOP);
 }
 
 void
 _gst_async_signal_and_unregister (OOP semaphoreOOP)
 {
-  _gst_disable_interrupts ();	/* block out everything! */
-  
-  queued_async_signals[async_queue_index].sem = semaphoreOOP;
-  queued_async_signals[async_queue_index++].unregister = true;
-  
-  SET_EXCEPT_FLAG (true);
-  _gst_enable_interrupts ();
+  _gst_disable_interrupts (false);	/* block out everything! */
+  _gst_async_call (do_async_signal, semaphoreOOP);
+  _gst_async_call (_gst_unregister_oop, semaphoreOOP);
+  _gst_enable_interrupts (false);
 }
 
 void
@@ -1630,8 +1630,6 @@ _gst_sync_wait (OOP semaphoreOOP)
 {
   gst_semaphore sem;
 
-  _gst_disable_interrupts ();	/* block out everything! */
-  
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
   if (TO_INT (sem->signals) <= 0)
     {
@@ -1647,7 +1645,6 @@ _gst_sync_wait (OOP semaphoreOOP)
     sem->signals = DECR_INT (sem->signals);
 
   /* printf ("wait %O %O\n", semaphoreOOP, sem->firstLink); */
-  _gst_enable_interrupts ();
 }
 
 OOP
@@ -1685,11 +1682,25 @@ resume_process (OOP processOOP,
 
   /* 2002-19-12: tried get_active_process instead of get_scheduled_process.  */
   activeOOP = get_active_process ();
-  if (processOOP == activeOOP)
-    return (true);
-
   active = (gst_process) OOP_TO_OBJ (activeOOP);
   process = (gst_process) OOP_TO_OBJ (processOOP);
+  priority = TO_INT (process->priority);
+
+  /* As a special exception, don't preempt a process that has disabled
+     interrupts. ### this behavior is currently disabled.  */
+  ints_enabled = IS_NIL (active->interrupts)
+	         || TO_INT(active->interrupts) <= 0;
+
+  /* resume_process is also used when changing the priority of a ready/active
+     process.  In this case, first remove the process from its current list.  */
+  if (processOOP == activeOOP)
+    {
+      assert (!alwaysPreempt);
+      remove_process_from_list (processOOP);
+    }
+  else if (priority >= TO_INT (active->priority) /* && ints_enabled */ )
+    alwaysPreempt = true;
+
   if (IS_NIL (processOOP) || is_process_terminating (processOOP))
     /* The process was terminated - nothing to resume, fail */
     return (false);
@@ -1701,17 +1712,10 @@ resume_process (OOP processOOP,
       return (true);
     }
 
-  /* As a special exception, don't preempt a process that has disabled
-     interrupts. ### this behavior is currently disabled.  */
-  ints_enabled = IS_NIL (active->interrupts)
-	         || TO_INT(active->interrupts) <= 0;
-
-  priority = TO_INT (process->priority);
   processLists = GET_PROCESS_LISTS ();
   processList = ARRAY_AT (processLists, priority);
 
-  if ((priority >= TO_INT (active->priority) /* && ints_enabled */ )
-      || alwaysPreempt)
+  if (alwaysPreempt)
     {
       /* We're resuming a process with a *equal or higher* priority, so sleep
          the current one and activate the new one */
@@ -1719,11 +1723,13 @@ resume_process (OOP processOOP,
       activate_process (processOOP);
     }
   else
-    /* this process has a lower priority than the active one, so the
-       policy is that it doesn't preempt the currently running one.
-       Anyway, it must be the first in its priority queue - so don't
-       put it to sleep.  */
-    add_first_link (processList, processOOP);
+    {
+      /* this process has a lower priority than the active one, so the
+         policy is that it doesn't preempt the currently running one.
+         Anyway, it must be the first in its priority queue - so don't
+         put it to sleep.  */
+      add_first_link (processList, processOOP);
+    }
 
   return (true);
 }
@@ -2106,6 +2112,10 @@ _gst_init_interpreter (void)
 
   _gst_this_context_oop = _gst_nil_oop;
   async_queue_index = 0;
+  async_queue_index_sig = 0;
+  async_queue_size = 32;
+  queued_async_signals = malloc (sizeof (async_queue_entry) * async_queue_size);
+  gl_lock_init (async_queue_lock);
 
   for (i = 0; i < MAX_LIFO_DEPTH; i++)
     lifo_contexts[i].flags = F_POOLED | F_CONTEXT;
@@ -2118,7 +2128,8 @@ OOP
 _gst_prepare_execution_environment (void)
 {
   gst_method_context dummyContext;
-  OOP dummyContextOOP;
+  OOP dummyContextOOP, processOOP;
+  inc_ptr inc = INC_SAVE_POINTER ();
 
   empty_context_stack ();
   dummyContext = alloc_stack_context (4);
@@ -2141,7 +2152,12 @@ _gst_prepare_execution_environment (void)
   dummyContextOOP = alloc_oop (dummyContext,
 			       _gst_mem.active_flag | F_POOLED | F_CONTEXT);
 
-  return (create_callin_process (dummyContextOOP));
+
+  INC_ADD_OOP (dummyContextOOP);
+  processOOP = create_callin_process (dummyContextOOP);
+
+  INC_RESTORE_POINTER (inc);
+  return (processOOP);
 }
 
 OOP
@@ -2176,7 +2192,9 @@ _gst_nvmsg_send (OOP receiver,
   for (i = 0; i < sendArgs; i++)
     PUSH_OOP (args[i]);
 
-  if (OOP_CLASS (sendSelector) == _gst_symbol_class)
+  if (!sendSelector)
+    send_block_value (sendArgs, sendArgs);
+  else if (OOP_CLASS (sendSelector) == _gst_symbol_class)
     SEND_MESSAGE (sendSelector, sendArgs);
   else
     _gst_send_method (sendSelector);
@@ -2193,14 +2211,9 @@ _gst_nvmsg_send (OOP receiver,
   /* Re-enable the previously executing process *now*, because a
      primitive might expect the current stack pointer to be that
      of the process that was executing.  */
-  if (reentrancy_jmp_buf && !--reentrancy_jmp_buf->suspended)
+  if (reentrancy_jmp_buf && !--reentrancy_jmp_buf->suspended
+      && !is_process_terminating (reentrancy_jmp_buf->processOOP))
     {
-      if (is_process_terminating (reentrancy_jmp_buf->processOOP))
-	{
-	  _gst_errorf ("Process terminated during call-out, VM confused!\n");
-	  abort ();
-	}
-
       resume_process (reentrancy_jmp_buf->processOOP, true);
       if (!IS_NIL (switch_to_process))
         change_process_context (switch_to_process);
@@ -2275,10 +2288,14 @@ copy_semaphore_oops (void)
 {
   int i;
 
-  _gst_disable_interrupts ();	/* block out everything! */
+  _gst_disable_interrupts (false);	/* block out everything! */
 
   for (i = 0; i < async_queue_index; i++)
-    MAYBE_COPY_OOP (queued_async_signals[i].sem);
+    if (queued_async_signals[i].data)
+      MAYBE_COPY_OOP (queued_async_signals[i].data);
+  for (i = 0; i < async_queue_index_sig; i++)
+    if (queued_async_signals_sig[i].data)
+      MAYBE_COPY_OOP (queued_async_signals_sig[i].data);
 
   /* there does seem to be a window where this is not valid */
   if (single_step_semaphore)
@@ -2287,7 +2304,7 @@ copy_semaphore_oops (void)
   /* there does seem to be a window where this is not valid */
   MAYBE_COPY_OOP (switch_to_process);
 
-  _gst_enable_interrupts ();
+  _gst_enable_interrupts (false);
 }
 
 
@@ -2309,10 +2326,14 @@ mark_semaphore_oops (void)
 {
   int i;
 
-  _gst_disable_interrupts ();	/* block out everything! */
+  _gst_disable_interrupts (false);	/* block out everything! */
 
   for (i = 0; i < async_queue_index; i++)
-    MAYBE_MARK_OOP (queued_async_signals[i].sem);
+    if (queued_async_signals[i].data)
+      MAYBE_MARK_OOP (queued_async_signals[i].data);
+  for (i = 0; i < async_queue_index_sig; i++)
+    if (queued_async_signals_sig[i].data)
+      MAYBE_MARK_OOP (queued_async_signals_sig[i].data);
 
   /* there does seem to be a window where this is not valid */
   if (single_step_semaphore)
@@ -2321,7 +2342,7 @@ mark_semaphore_oops (void)
   /* there does seem to be a window where this is not valid */
   MAYBE_MARK_OOP (switch_to_process);
 
-  _gst_enable_interrupts ();
+  _gst_enable_interrupts (false);
 }
 
 
@@ -2406,80 +2427,31 @@ _gst_restore_object_pointers (void)
   SET_EXCEPT_FLAG (true);	/* force to import registers */
 }
 
-void
-_gst_init_signals (void)
+static RETSIGTYPE
+interrupt_on_signal (int sig)
 {
-  if (!_gst_make_core_file)
-    {
-#ifdef ENABLE_JIT_TRANSLATION
-      _gst_set_signal_handler (SIGILL, interrupt_handler);
-#endif
-      _gst_set_signal_handler (SIGABRT, interrupt_handler);
-    }
-  _gst_set_signal_handler (SIGTERM, interrupt_handler);
-  _gst_set_signal_handler (SIGINT, interrupt_handler);
-  _gst_set_signal_handler (SIGFPE, interrupt_handler);
-  _gst_set_signal_handler (SIGUSR1, interrupt_handler);
-}
-
-
-void
-stop_executing (const char *msg)
-{
-  _gst_abort_execution = msg;
-  SET_EXCEPT_FLAG (true);
   if (reentrancy_jmp_buf)
-    longjmp (reentrancy_jmp_buf->jmpBuf, 1);	/* throw out from C
-						   code */
+    stop_execution ();
+  else
+    {
+      _gst_set_signal_handler (sig, SIG_DFL);
+      raise (sig);
+    }
 }
 
-
-RETSIGTYPE
-interrupt_handler (int sig)
+static void
+backtrace_on_signal_1 (mst_Boolean is_serious_error, mst_Boolean c_backtrace)
 {
-  mst_Boolean is_serious_error = true;
-  mst_Boolean in_c_code = !in_interpreter || !ip || _gst_gc_running;
+  static int reentering = -1;
 
-  switch (sig)
-    {
-    case SIGTERM:
-      is_serious_error = false;
-      break;
+  /* Avoid recursive signals */
+  reentering++;
 
-    case SIGUSR1:
-      is_serious_error = false;
-      _gst_set_signal_handler (sig, interrupt_handler);
-      break;
-
-    case SIGFPE:
-      _gst_set_signal_handler (sig, interrupt_handler);
-      return;
-
-    case SIGINT:
-      is_serious_error = false;
-      if (!_gst_non_interactive && in_interpreter)
-	{
-	  _gst_set_signal_handler (sig, interrupt_handler);
-	  stop_executing ("userInterrupt");
-	  return;
-	}
-      break;
-
-    default:
-      break;
-    }
-
-  if (sig != SIGUSR1)
-    _gst_errorf ("%s", strsignal (sig));
-
-  if (!in_c_code)
-    {
-      /* Avoid recursive signals */
-      mst_Boolean save_in_interpreter = in_interpreter;
-      in_interpreter = false;
-      _gst_show_backtrace ();
-      in_interpreter = save_in_interpreter;
-    }
+  if ((reentrancy_jmp_buf && reentrancy_jmp_buf->interpreter)
+      && !reentering
+      && ip
+      && !_gst_gc_running)
+    _gst_show_backtrace ();
   else
     {
       if (is_serious_error)
@@ -2488,8 +2460,7 @@ interrupt_handler (int sig)
 #ifdef HAVE_EXECINFO_H
       /* Don't print a backtrace, for example, if exiting during a
 	 compilation.  */
-      if ((_gst_verbosity == 3 && (ip || _gst_gc_running))
-	  || is_serious_error || sig == SIGUSR1)
+      if (c_backtrace && !reentering)
 	{
           PTR array[11];
           size_t size = backtrace (array, 11);
@@ -2498,21 +2469,42 @@ interrupt_handler (int sig)
 #endif
     }
 
-  switch (sig)
-    {
-    case SIGUSR1:
-      return;
-
-    case SIGTERM:
-    case SIGINT:
-      exit (0);
-
-    default:
-      _gst_set_signal_handler (sig, SIG_DFL);
-      raise (sig);
-    }
+  reentering--;
 }
 
+static RETSIGTYPE
+backtrace_on_signal (int sig)
+{
+  _gst_errorf ("%s", strsignal (sig));
+  _gst_set_signal_handler (sig, backtrace_on_signal);
+  backtrace_on_signal_1 (sig != SIGTERM, sig != SIGTERM);
+  _gst_set_signal_handler (sig, SIG_DFL);
+  raise (sig);
+}
+
+static RETSIGTYPE
+user_backtrace_on_signal (int sig)
+{
+  _gst_set_signal_handler (sig, user_backtrace_on_signal);
+  backtrace_on_signal_1 (false, true);
+}
+
+void
+_gst_init_signals (void)
+{
+  if (!_gst_make_core_file)
+    {
+#ifdef ENABLE_JIT_TRANSLATION
+      _gst_set_signal_handler (SIGILL, backtrace_on_signal);
+#endif
+      _gst_set_signal_handler (SIGABRT, backtrace_on_signal);
+    }
+  _gst_set_signal_handler (SIGTERM, backtrace_on_signal);
+  _gst_set_signal_handler (SIGINT, interrupt_on_signal);
+  _gst_set_signal_handler (SIGUSR1, user_backtrace_on_signal);
+}
+
+
 void
 _gst_show_backtrace (void)
 {
@@ -2542,7 +2534,7 @@ _gst_show_backtrace (void)
 	      if (IS_NIL(context->parentContext))
 	        printf ("<bottom>\n");
 	      else
-	        printf ("<call-in>\n");
+	        printf ("<unwind point>\n");
 	      continue;
 	    }
 
@@ -2659,10 +2651,55 @@ _gst_set_primitive_attributes (int primitive, prim_table_entry *pte)
 }
 
 void
+push_jmp_buf (interp_jmp_buf *jb, int for_interpreter, OOP processOOP)
+{
+  jb->next = reentrancy_jmp_buf;
+  jb->processOOP = processOOP;
+  jb->suspended = 0;
+  jb->interpreter = for_interpreter;
+  jb->interrupted = false;
+  _gst_register_oop (processOOP);
+  reentrancy_jmp_buf = jb;
+}
+
+mst_Boolean
+pop_jmp_buf (void)
+{
+  interp_jmp_buf *jb = reentrancy_jmp_buf;
+  reentrancy_jmp_buf = jb->next;
+
+  if (jb->interpreter && !is_process_terminating (jb->processOOP))
+    _gst_terminate_process (jb->processOOP);
+    
+  _gst_unregister_oop (jb->processOOP);
+  return jb->interrupted && reentrancy_jmp_buf;
+}
+
+void
+stop_execution (void)
+{
+  reentrancy_jmp_buf->interrupted = true;
+
+  if (reentrancy_jmp_buf->interpreter
+      && !is_process_terminating (reentrancy_jmp_buf->processOOP))
+    {
+      _gst_abort_execution = "userInterrupt";
+      SET_EXCEPT_FLAG (true);
+      if (get_active_process () != reentrancy_jmp_buf->processOOP)
+	resume_process (reentrancy_jmp_buf->processOOP, true);
+    }
+  else
+    longjmp (reentrancy_jmp_buf->jmpBuf, 1);
+}
+
+mst_Boolean
 parse_stream_with_protection (mst_Boolean method)
 {
-  interp_jmp_buf localJmpBuf;
+  interp_jmp_buf jb;
 
-  PROTECT_CURRENT_PROCESS_WITH (&localJmpBuf)
+  push_jmp_buf (&jb, false, get_active_process ());
+  if (setjmp (jb.jmpBuf) == 0)
     _gst_parse_stream (method);
+
+  return pop_jmp_buf ();
 }
