@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008
  * Free Software Foundation, Inc.
  * Written by Steve Byrne and Paolo Bonzini.
  *
@@ -54,10 +54,6 @@
 
 #include "gstpriv.h"
 
-#ifdef HAVE_SIGSEGV_H
-#include "sigsegv.h"
-#endif
-
 #define	K		1024
 #define INIT_NUM_INCUBATOR_OOPS   50
 #define INCUBATOR_CHUNK_SIZE	  20
@@ -74,18 +70,9 @@
 /* Define this flag to disable incremental garbage collection */
 /* #define NO_INCREMENTAL_GC */
 
-/* Define this flag to disable blacking of grey pages (that is, the
-   entire oldspace is scanned to look for reachable newspace objects).
-   This is also necessary to run valgrind on GNU Smalltalk.  */
-/* #define NO_SIGSEGV_HANDLING */
-
 /* Define this flag to turn on debugging code for oldspace management */
 /* #define MMAN_DEBUG_OUTPUT */
 
-#if !defined HAVE_SIGSEGV_RECOVERY || !(HAVE_SIGSEGV_RECOVERY-0)
-#define NO_SIGSEGV_HANDLING
-#endif
-
 #if defined(GC_DEBUG_OUTPUT)
 #define GC_DEBUGGING
 #endif
@@ -637,8 +624,6 @@ _gst_make_oop_weak (OOP oop)
   weak_area_tree *node = NULL;
   rb_node_t **p = (rb_node_t **) &_gst_mem.weak_areas;
 
-  /* Weak OOPs must be fixed.  */
-  _gst_make_oop_fixed (oop);
   oop->flags |= F_WEAK;
   _gst_mem.numWeakOOPs++;
 
@@ -651,16 +636,10 @@ _gst_make_oop_weak (OOP oop)
       else if (oop > node->oop)
         p = &(*p)->rb_right;
       else
-        {
-	  node->base = (OOP *) oop->object;
-	  node->end = (OOP *) oop->object + TO_INT (oop->object->objSize);
-	  return;
-	}
+	return;
     }
 
   entry = (weak_area_tree *) xmalloc (sizeof (weak_area_tree));
-  entry->base = (OOP *) oop->object;
-  entry->end = (OOP *) oop->object + TO_INT (oop->object->objSize);
   entry->oop = oop;
   entry->rb.rb_parent = &node->rb;
   entry->rb.rb_left = entry->rb.rb_right = NULL;
@@ -1167,6 +1146,7 @@ _gst_global_gc (int next_allocation)
   if (!_gst_mem.old->heap_limit)
     _gst_mem.old->heap_limit = old_limit;
 
+  _gst_invalidate_croutine_cache ();
   mourn_objects ();
 }
 
@@ -1248,6 +1228,7 @@ _gst_scavenge (void)
     _gst_mem.factor * tenuredBytes +
     (1 - _gst_mem.factor) * _gst_mem.tenuredBytesPerScavenge;
 
+  _gst_invalidate_croutine_cache ();
   mourn_objects ();
 }
 
@@ -1648,9 +1629,6 @@ _gst_tenure_all_survivors ()
 void
 check_weak_refs ()
 {
-  OOP oop = NULL, *field;
-  mst_Boolean mourn;
-
   rb_node_t *node;
   rb_traverse_t t;
 
@@ -1658,11 +1636,18 @@ check_weak_refs ()
        node; node = rb_next (&t))
     {
       weak_area_tree *area = (weak_area_tree *) node;
+      mst_Boolean mourn = false;
+      OOP *field, oop;
+      int n;
 
-      mourn = false;
-      for (field = area->base; field < area->end; field++)
+      oop = area->oop;
+      if (!IS_OOP_VALID_GC (oop))
+	continue;
+
+      for (field = (OOP *) oop->object, n = TO_INT (oop->object->objSize);
+	   n--; field++)
         {
-	  oop = *field;
+	  OOP oop = *field;
           if (IS_INT (oop))
 	    continue;
 
@@ -1674,7 +1659,7 @@ check_weak_refs ()
         }
 
       if (mourn)
-        _gst_add_buf_pointer (oop);
+        _gst_add_buf_pointer (area->oop);
     }
 }
 
@@ -1766,7 +1751,12 @@ scan_grey_pages ()
   for (last = NULL, next = &_gst_mem.grey_pages.head; (node = *next); )
     {
 #if defined(GC_DEBUG_OUTPUT) || defined(MMAN_DEBUG_OUTPUT)
-      printf ("Scanning grey page %p...%p\n", node->base, node->base + node->n);
+      printf ("Scanning grey page %p...%p ", node->base, node->base + node->n);
+#if defined(GC_DEBUG_OUTPUT)
+      putchar ('\n');
+#else
+      fflush (stdout);
+#endif
 #endif
 
       PREFETCH_START (node->base, PREF_READ | PREF_NTA);
@@ -1803,6 +1793,9 @@ scan_grey_pages ()
       else
 #endif
 	{
+#if defined (MMAN_DEBUG_OUTPUT)
+	  printf ("Found %d pointers\n", n);
+#endif
 	  last = node;
 	  next = &(node->next);
 	}
@@ -2209,14 +2202,14 @@ _gst_mark_an_oop_internal (OOP oop,
 	    }
 	  else if UNCOMMON (oop->flags & (F_EPHEMERON | F_WEAK))
 	    {
+	      if (oop->flags & F_EPHEMERON)
+	        _gst_add_buf_pointer (oop);
+
 	      /* In general, there will be many instances of a class,
 		 but only the first time will it be unmarked.  So I'm
 		 marking this as uncommon.  */
 	      if UNCOMMON (!IS_OOP_MARKED (objClass))
 		TAIL_MARK_OOP (objClass);
-
-	      if (oop->flags & F_EPHEMERON)
-	        _gst_add_buf_pointer (oop);
 	    }
 	  else
 	    {
